
#define PROFILER_IMPLEMENTATION
// #define PROFILER_ON
// #define PROFILER_ASSERT
// #define PROFILER_DEBUG
#define PROFILER_MAX_COLLAPSABLE_COUNT 360
#include "profiler_ids.h"
#include "../lib/profiler.h"

#include "../lib/common.h"
#include "../lib/gl_hardware.h"
#include "../lib/window.h"
#include "../lib/raster_font.h"
#include "../lib/image.h"
#include "../lib/tga.h"

void render_clear( u8* buffer, s32 width, s32 height, s32 pitch ) {

    profiler_event_start( render_clear );

    u32* pixels = ( u32* ) buffer;

    for ( umm y = 0; y < height; y++ ) {
        for ( umm x = 0; x < width; x++ ) {
            pixels[ x ] = 0x00000000;
        }

        pixels += pitch;
    }

    profiler_event_end( render_clear );
}

void render_texture_rgba( u8* frame_buffer, s32 frame_buffer_width, s32 frame_buffer_height, s32 frame_buffer_pitch, u8* texture, s32 texture_width, s32 texture_height, s32 texture_pitch, s32 texture_x, s32 texture_y ) {

    profiler_event_start( render_texture_rgba );

    rect2i frame_buffer_rect = r2i_min_size( 0, 0, frame_buffer_width, frame_buffer_height );
    rect2i texture_rect = r2i_min_size( texture_x, texture_y, texture_width, texture_height );

    rect2i to_copy = r2i_intersection( frame_buffer_rect, texture_rect );

    s32 x_offset = to_copy.left - texture_x;
    s32 y_offset = to_copy.bottom - texture_y;

    u8* destination = frame_buffer + ( texture_y * frame_buffer_pitch + to_copy.left ) * 4;
    u8* source = texture + ( y_offset * texture_pitch + x_offset ) * 4;

    for ( smm y = to_copy.bottom; y < to_copy.top; y++ ) {

        u8* destination_pixel = destination;
        u8* source_pixel = source;

        for ( smm x = to_copy.left; x < to_copy.right; x++ ) {

            u8 d_r = destination_pixel[ 0 ];
            u8 d_g = destination_pixel[ 1 ];
            u8 d_b = destination_pixel[ 2 ];
            u8 d_a = destination_pixel[ 3 ];

            u8 s_r = source_pixel[ 0 ];
            u8 s_g = source_pixel[ 1 ];
            u8 s_b = source_pixel[ 2 ];
            u8 s_a = source_pixel[ 3 ];

            u8 inv_s_a = 255 - s_a;

            u8 r = ( inv_s_a * d_r + ( s_r * s_a ) ) / 255;
            u8 g = ( inv_s_a * d_g + ( s_g * s_a ) ) / 255;
            u8 b = ( inv_s_a * d_b + ( s_b * s_a ) ) / 255;
            u8 a = 255;

            destination_pixel[ 0 ] = r;
            destination_pixel[ 1 ] = g;
            destination_pixel[ 2 ] = b;
            destination_pixel[ 3 ] = a;

            destination_pixel += 4;
            source_pixel += 4;
        }

        destination += frame_buffer_pitch * 4;
        source += texture_pitch * 4;
    }

    profiler_event_end( render_texture_rgba );
}


u8 batch_id_textured = 0xff;
u8 batch_id_ascii = 0xff;
u8 batch_id_rectangle = 0xff;

b32 g_render_as_ascii = false;

data_t texture_names[ ] = {
    string_l( "mug_01.tga" ),
    string_l( "mug_02.tga" ),
    string_l( "mug_03.tga" ),
    string_l( "boat.tga" ),
    string_l( "sugar_under.tga" ),
    string_l( "sugar_over.tga" ),
    string_l( "sugar_wave_01.tga" ),
    string_l( "sugar_wave_02.tga" ),
    string_l( "boat_64_01.tga" ),
    string_l( "boat_64_02.tga" ),
    string_l( "boat_64_03.tga" ),
    string_l( "boat_64_04.tga" ),
    string_l( "boat_64_05.tga" ),
    string_l( "boat_ascii.tga" ),
    string_l( "squid_idle_ascii_right.tga" ),
    string_l( "squid_idle_ascii_left.tga" ),
    string_l( "squid_move_ascii_right.tga" ),
    string_l( "squid_move_ascii_left.tga" ),
};

typedef enum texture_id_t {

    texture_id_mug_01,
    texture_id_mug_02,
    texture_id_mug_03,
    texture_id_boat,
    texture_id_sugar_under,
    texture_id_sugar_over,
    texture_id_sugar_wave_01,
    texture_id_sugar_wave_02,
    texture_id_boat_64_01,
    texture_id_boat_64_02,
    texture_id_boat_64_03,
    texture_id_boat_64_04,
    texture_id_boat_64_05,
    texture_id_boat_ascii,
    texture_id_squid_idle_ascii_right,
    texture_id_squid_idle_ascii_left,
    texture_id_squid_move_ascii_right,
    texture_id_squid_move_ascii_left,

} texture_id_t;

gl_texture_t g_textures[ array_count( texture_names ) ] = { 0 };

typedef enum chapter_t {
    chapter_1,
    chapter_2,
    chapter_3,
    chapter_4,
    chapter_5,
} chapter_t;

typedef struct game_state_t {

    chapter_t chapter;
    b32 initialized;
    u64 seed;

    struct {
        r32 x, y;
        r32 scale;
        umm image_index;
        r32 blink_time;
    } c1;

    struct {
        r32 x, y;
        r32 sugar_target[ 10 ];
        vec2 sugar_positions[ 10 ];
        r32 sugar_timer[ 10 ];
        u32 next_sugar;
        r32 time;
    } c2;

    struct {
        r32 x, y;
        r32 time;
        r32 scale;
        b32 started;
        vec2 particle_positions[ 25 ];
        u32 particle_count;
        r32 particle_time;
        r32 particle_time_min;
        r32 particle_time_max;
        r32 switch_timer;
    } c3;

    struct {
        r32 x, y;
        r32 scale;
        r32 target_x;
        r32 target_y;
        r32 target_scale;
        r32 blink_time;
    } c4;

    struct {
        r32 x, y;
        b32 just_switched;
        b32 use_left;
        u32 screen_index;
    } c5;

} game_state_t;

data_t software_framebuffer = { 0 };

#include "game.c"

int main( int argc, char** argv ) {

    arguments_initialize( argc, argv );

    u32 error = 0;
    profiler_initialize( main, 100000000, &error );

    window_state_t window = { 0 };
    window_create( &window, "Le bateau ivre", 1280, 720, s32_max, s32_max, window_create_flag_no_title_bar );

    for ( umm i = 0; i < array_count( texture_names ); i++ ) {

        data_t file = file_read_entire( texture_names[ i ], 0 );
        u8* pixels = file.bytes + 18;
        u32 width = tga_get_width( file.bytes );
        u32 height = tga_get_height( file.bytes );
        u8 bytes_per_pixels = tga_get_bytes_per_pixel( file.bytes );
        g_textures[ i ] = gl_create_texture( pixels, width, height, bytes_per_pixels, false );
        // file_free_memory( &file );
    }

    gl_state_t gl = gl_start( window.width, window.height, gl_state_flag_alpha_blending );
    batch_id_textured = gl_add_textured_quad( &gl );
    u32 shader_handle = gl_create_shader_program( vs_screen_space_texture_color, fs_default_u8_color );
    gl_texture_t atlas_texture = gl_create_texture( raster_font, raster_font_width, raster_font_height, 1, 0 );
    batch_id_ascii = gl_add_atlas_extended( &gl, 50000, 50000, 0, shader_handle, 0, atlas_texture.handle );
    gl_batch_t* atlas = gl.batches + batch_id_ascii;

    batch_id_rectangle = gl_add_rectangle( &gl );

    umm size = window.width * window.height;

    software_framebuffer = memory_get_virtual( size * 4, 0 );
    data_t gray_scale = memory_get_virtual( size, 0 );

    image_to_ascii_t ascii = {
        raster_font,
        ( u16 ) raster_font_width,
        ( u16 ) raster_font_height,
        ( u16 ) raster_font_width,
        0x20, 0x7e,
        // 0, 0xfe,
        16, 16,
        8, 12
    };

    image_to_ascii_initialize( &ascii );

    u32 max_x_character_count = window.width / ascii.element_width;
    u32 max_y_character_count = window.height / ascii.element_height;

    data_t character_buffer = memory_get_virtual( max_x_character_count * max_y_character_count, 0 );

    gl_texture_t framebuffer = gl_create_texture( software_framebuffer.bytes, window.width, window.height, 4, 0 );

    game_state_t game_state = { 0 };

    u64 frequency, last_time = 0;
    perf_get_frequency( &frequency );
    perf_get_time( &last_time );

    game_state.seed = last_time;

    while ( window.running ) {

        profiler_collapsable_start( frame_loop );
        r32 delta_time = 1.0f / 60.0f;

        u64 time = last_time;
        perf_get_time( &last_time );
        delta_time = ( r32 ) ( ( r64 ) ( last_time - time ) / ( r64 ) frequency );

        gl_frame_start( &gl );

        window_handle_messages( &window, 0 );

        if ( window.close_requested ) {
            window.running = false;
        }

        if ( window.resized ) {
            profiler_event_start( resize );
            gl_resize_viewport( &gl, 0, 0, window.width, window.height );
            gl_free_texture( &framebuffer );
            framebuffer = gl_create_texture( software_framebuffer.bytes, window.width, window.height, 4, 0 );
            profiler_event_end( resize );
        }

        if ( input_key_just_pressed( &window, vk_f1 ) ) {
            g_extra_step = !g_extra_step;
        }

        if ( input_key_just_pressed( &window, vk_escape ) ) {
            window.running = false;
        }

        gl_clear( &gl.commands, v4( 0, 0, 0, 1 ) );
        render_clear( software_framebuffer.bytes, window.width, window.height, window.width );

        game_update( &game_state, delta_time, &gl, &window );

        if ( g_render_as_ascii ) {

            profiler_event_start( to_grayscale );
            image_rgba_to_grayscale( software_framebuffer.bytes, gray_scale.bytes, window.width, window.height, window.width );
            profiler_event_end( to_grayscale );

            umm x_character_count, y_character_count;
            image_to_ascii_render( gray_scale.bytes, window.width, window.height, window.width, &ascii, character_buffer.bytes, character_buffer.reserved, &x_character_count, &y_character_count );

            gl_frame_start( &gl );

            gl_clear( &gl.commands, v4( 0, 0, 0, 1 ) );

            profiler_event_start( batch_characters );

            u32 triangle_count = 2;
            u8 vertex_size = gl_vertex_format_compute_vertex_size( atlas->vertex_format );
            umm va_size = ( 2 + triangle_count ) * vertex_size * sizeof( r32 );

            r32 u_width = 1.0f / ascii.x_element_count;
            r32 v_height = 1.0f / ascii.y_element_count;

            for ( u16 y = 0; y < y_character_count; y++ ) {

                rect2 character_rect = r2_min_size( 0, ( r32 ) ( y * ascii.element_height ), 0, ( r32 ) ascii.element_height );

                for ( u16 x = 0; x < x_character_count; x++ ) {

                    umm c = character_buffer.bytes[ y * x_character_count + x ];
                    umm cx = c % ascii.x_element_count;
                    umm cy = c / ascii.x_element_count;

                    character_rect.left = ( r32) ( x * ascii.element_width );
                    character_rect.right = character_rect.left + ascii.element_width;

                    r32 uvs[ 4 ] = {
                        cx * u_width, cy * v_height,
                        ( cx + 1 ) * u_width, ( cy + 1 ) * v_height
                    };

                    // vec4 color = g_extra_step ? yellow : white;

                    gl_command_t* result = memory_push_struct( &atlas->commands, gl_command_t );
                    result->type = gl_command_type_triangle_strip;
                    atlas->command_count++;
                    result->triangle_strip.vertex_array = ( r32* ) memory_push_size( &atlas->datas, va_size, 4 );
                    result->triangle_strip.triangle_count = 2;
                    r32* va = result->triangle_strip.vertex_array;
                    u32* va_u32 = ( u32* ) va;

                    va[ 0 ] = character_rect.e[ 0 ];
                    va[ 1 ] = character_rect.e[ 3 ];
                    va_u32[ 2 ] = 0xffffffff;
                    va[ 3 ] = uvs[ 0 ];
                    va[ 4 ] = uvs[ 3 ];

                    va[ 5 ] = character_rect.e[ 0 ];
                    va[ 6 ] = character_rect.e[ 1 ];
                    va_u32[ 7 ] = 0xffffffff;
                    va[ 8 ] = uvs[ 0 ];
                    va[ 9 ] = uvs[ 1 ];

                    va[ 10 ] = character_rect.e[ 2 ];
                    va[ 11 ] = character_rect.e[ 3 ];
                    va_u32[ 12 ] = 0xffffffff;
                    va[ 13 ] = uvs[ 2 ];
                    va[ 14 ] = uvs[ 3 ];

                    va[ 15 ] = character_rect.e[ 2 ];
                    va[ 16 ] = character_rect.e[ 1 ];
                    va_u32[ 17 ] = 0xffffffff;
                    va[ 18 ] = uvs[ 2 ];
                    va[ 19 ] = uvs[ 1 ];

                    // gl_batch_rectangle( atlas, character_rect.e, uvs, 0, color );
                }
            }

            profiler_event_end( batch_characters );

            profiler_event_start( render );
            gl_render_state( &gl );
            profiler_event_end( render );

        } else {

            gl_render_state( &gl );
        }

        window_swap_buffers( &window );

        profiler_collapsable_end( frame_loop );
    }

    profiler_finalize( );

    if ( arguments_find_l( "-p", 0 ) ) {
        profiler_write_to_file( "profile.bin", true, &error );
    }

    return 0;
}