
#ifndef GL_H

#include "font.h"

#include "shaders.h"

/* TODO simon: redo the gl_command_rectangle_t like structure, to avoid separationg command and data ? Having command followed by data seem that it would work ? Not sure, because how do we define the size of the memory required without passing command count and primitive count ? */

typedef enum gl_command_type_t {
    
    gl_command_type_none,
    
    gl_command_type_clear,
    gl_command_type_viewport,
    gl_command_type_point,
    gl_command_type_line,
    gl_command_type_triangle_strip,
    gl_command_type_triangle_strip_textured,
    gl_command_type_text,
    gl_command_type_batch,
    gl_command_type_scissor,
    
    gl_command_type_count
        
} gl_command_type_t;

typedef struct gl_batch_t {
    
    gl_command_type_t type;
    u32 command_count;
    data_t commands;
    data_t datas;
    u32 vertex_handle;
    u32 v_buffer_size;
    u32 vertex_used;
    u32 vertex_index;
    u32 index_handle;
    u32 i_buffer_size;
    u32 index_used;
    u32 shader_handle;
    r32* matrix;
    u8 vertex_format; /* TODO simon: should I just change the vertex format to be a simple structure ? */
    font_collection_t* collection;
    u32 texture;
    
} gl_batch_t;

typedef struct gl_command_t {
    
    gl_command_type_t type;
    
    union {
        
        struct {
            u32 color;
        } clear;
        
        struct {
            vec2 position;
            vec2 size;
        } viewport;
        
        struct {
            b32 enable;
            s32 x, y, width, height;
        } scissor;
        
        struct {
            r32* vertex_array;
            u32 point_count;
            r32 size;
            u32 color;
        } point;
        
        struct {
            r32* vertex_array;
            u32 point_count;
            u32 color;
        } line;
        
        struct {
            r32* vertex_array;
            u32 triangle_count;
            u32 color;
        } triangle_strip;
        
        struct {
            r32* vertex_array;
            u32 triangle_count;
            u32 color;
            u32 texture;
            u32 bytes_per_pixel;
        } triangle_strip_textured;
        
        struct {
            vec3 position;
            u32 font_index;
            data_t string;
            u32 color;
            u32 codepoint_count;
            u32 outputed_codepoint_count;
            b32 do_first_line_character_offset;
        } text;
        
        struct {
            gl_batch_t* batch;
        } batch;
    };
} gl_command_t;

#define gl_command_t_alignment 8

typedef struct gl_command_buffer_t {
    
    u32 command_count;
    data_t commands;
    data_t datas;
} gl_command_buffer_t;

typedef struct gl_texture_t {
    
    u8* pixels;
    u32 handle;
    u32 width;
    u32 height;
    u8 bytes_per_pixel;
    
} gl_texture_t;

u8 const gl_primitive_restart_index_u8 = u8_max;
u16 const gl_primitive_restart_index_u16 = u16_max;
u32 const gl_primitive_restart_index_u32 = u32_max;

/* MARK vertex_format */

u8 gl_vertex_format_position_mask = 0x03;
u8 gl_vertex_format_color_mask = 0x04;
u8 gl_vertex_format_uv_mask = 0x08;
u8 gl_vertex_format_index_mask = 0x30;

#define gl_vertex_format_get_position_component_count( format ) ( ( ( format ) & gl_vertex_format_position_mask ) + 2 )
#define gl_vertex_format_has_color( format ) ( ( ( format ) & gl_vertex_format_color_mask ) ? 1 : 0 )
#define gl_vertex_format_has_uv( format ) ( ( ( format ) & gl_vertex_format_uv_mask ) ? 1 : 0 )
#define gl_vertex_format_get_index_size_byte( format ) ( ( ( ( format ) & gl_vertex_format_index_mask ) == 0x30 ) ? 4 : ( ( ( format ) & gl_vertex_format_index_mask ) >> 4 ) )

u8 gl_vertex_format_create( umm position, umm color, umm uv, umm index ) {
    
    u8 result = ( u8 ) position - 2;
    
    if ( color ) {
        result |= gl_vertex_format_color_mask;
    }
    
    if ( uv ) {
        result |= gl_vertex_format_uv_mask;
    }
    
    if ( index <= 2 ) {
        result |= ( index << 4 );
    } else {
        _assert( index == 4 );
        result |= ( 3 << 4 );
    }
    
    return result;
}

u8 gl_vertex_format_compute_vertex_size( u8 vertex_format ) {
    u8 result = gl_vertex_format_get_position_component_count( vertex_format );
    result += ( gl_vertex_format_has_color( vertex_format ) ? 1 : 0 );
    result += ( gl_vertex_format_has_uv( vertex_format ) ? 2 : 0 );
    return result;
}

u8 gl_vertex_format_compute_vertex_size_byte( u8 vertex_format ) {
    u8 result = gl_vertex_format_compute_vertex_size( vertex_format );
    result *= sizeof( r32 );
    return result;
}

// MARK command

stu gl_command_buffer_t gl_create_command_buffer( data_t* memory, umm size, umm data_size ) {
    
    gl_command_buffer_t result = { 0 };
    result.command_count = 0;
    result.commands = memory_get_block( memory, size, 0 );
    result.datas = memory_get_block( memory, data_size, 0 );
    
    return result;
}

stu void gl_reset_command_buffer( gl_command_buffer_t* buffer ) {
    
    buffer->command_count = 0;
    buffer->commands.used = 0;
    buffer->datas.used = 0;
}

stu gl_command_t* gl_clear( gl_command_buffer_t* buffer, vec4 color ) {
    
    gl_command_t* result = memory_push_struct( &buffer->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        buffer->command_count++;
        result->type = gl_command_type_clear;
        result->clear.color = v4_to_u32_color( color );
    }
    
    return result;
}

stu gl_command_t* gl_viewport( gl_command_buffer_t* buffer, vec2 position, vec2 size ) {
    
    gl_command_t* result = memory_push_struct( &buffer->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        buffer->command_count++;
        result->type = gl_command_type_viewport;
        result->viewport.position = position;
        result->viewport.size = size;
    }
    
    return result;
}

stu gl_command_t* gl_scissor( gl_command_buffer_t* buffer, b32 enable, s32 x, s32 y, s32 width, s32 height ) {
    
    gl_command_t* result = memory_push_struct( &buffer->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        buffer->command_count++;
        result->type = gl_command_type_scissor;
        result->scissor.enable = enable;
        result->scissor.x = math_max( 0, x );
        result->scissor.y = math_max( 0, y );
        result->scissor.width = math_max( 0, width );
        result->scissor.height = math_max( 0, height );
    }
    
    return result;
}

#define gl_scissor_disable( buffer ) gl_scissor( ( buffer ), false, 0, 0, 0, 0 )

// MARK batch
stu gl_command_t* gl_batch_scissor( gl_batch_t* batch, b32 enable, s32 x, s32 y, s32 width, s32 height ) {
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_scissor;
        result->scissor.enable = enable;
        result->scissor.x = math_max( 0, x );
        result->scissor.y = math_max( 0, y );
        result->scissor.width = math_max( 0, width );
        result->scissor.height = math_max( 0, height );
    }
    
    return result;
}

#define gl_batch_scissor_rect( batch, rect ) gl_batch_scissor( ( batch ), true, \
( s32 ) ( ( rect ).left ), \
( s32 ) ( ( rect ).bottom ), \
( s32 ) ( ( rect ).right - ( rect ).left ), \
( s32 ) ( ( rect ).top - ( rect ).bottom ) )

#define gl_batch_scissor_disable( batch ) gl_batch_scissor( ( batch ), false, 0, 0, 0, 0 )

stu gl_command_t* gl_batch_points( gl_batch_t* batch, r32* vertex_array, u32 point_count, r32 size, vec4 color ) {
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_point;
        
        u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( batch->vertex_format );
        umm array_size = point_count * vertex_size_byte;
        result->point.vertex_array = ( r32* ) memory_push_copy_p( &batch->datas, vertex_array, array_size, 4 );
        _assert( result->point.vertex_array );
        
        result->point.point_count = point_count;
        result->point.size = size;
        result->point.color = v4_to_u32_color( color );
    }
    
    return result;
}

stu gl_command_t* gl_batch_lines( gl_batch_t* batch, r32* vertex_array, u32 point_count, vec4 color ) {
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_line;
        
        u8 vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
        umm array_size = point_count * vertex_size * sizeof( r32 );
        result->line.vertex_array = ( r32* ) memory_push_copy_p( &batch->datas, vertex_array, array_size, 4 );
        _assert( result->line.vertex_array );
        
        result->line.point_count = point_count;
        result->line.color = v4_to_u32_color( color );
    }
    
    return result;
}

stu gl_command_t* gl_batch_rectangle( gl_batch_t* batch, r32* rectangle, r32* uvs, vec4* colors, vec4 color ) {
    
    profiler_event_start( gl_batch_rectangle );
    _assert( rectangle );
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_triangle_strip;
        
        u32 triangle_count = 2;
        u8 vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
        umm array_size = ( 2 + triangle_count ) * vertex_size * sizeof( r32 );
        result->triangle_strip.vertex_array = ( r32* ) memory_push_size( &batch->datas, array_size, 4 );
        _assert( result->triangle_strip.vertex_array );
        result->triangle_strip.triangle_count = triangle_count;
        
        vec4 dummy_colors[ 4 ] = { white, white, white, white };
        r32 dummy_uvs[ 4 ] = { 0, 0, 1, 1 };
        
        r32* va = result->triangle_strip.vertex_array;
        u32* va_u32 = ( u32* ) va;
        
        b32 has_color = gl_vertex_format_has_color( batch->vertex_format );
        b32 has_uv = gl_vertex_format_has_uv( batch->vertex_format );
        
        if ( !colors ) {
            colors = dummy_colors;
        }
        
        if ( !uvs ) {
            uvs = dummy_uvs;
        }
        
        {
            umm index = 0;
            va[ index + 0 ] = rectangle[ 0 ];
            va[ index + 1 ] = rectangle[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 0 ];
            va[ index + 1 ] = rectangle[ 1 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 2 ];
            va[ index + 1 ] = rectangle[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 2 ];
            va[ index + 1 ] = rectangle[ 1 ];
        }
        
        if ( has_color ) {
            umm index = 2;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 2 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 0 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 3 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 1 ], color ) );
        }
        
        if ( has_uv ) {
            umm index = 2;
            if ( has_color ) {
                index = 3;
            }
            va[ index + 0 ] = uvs[ 0 ];
            va[ index + 1 ] = uvs[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 0 ];
            va[ index + 1 ] = uvs[ 1 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 2 ];
            va[ index + 1 ] = uvs[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 2 ];
            va[ index + 1 ] = uvs[ 1 ];
        }
    }
    
    profiler_event_end( gl_batch_rectangle );
    return result;
}

stu gl_command_t* gl_batch_rectangle_textured( gl_batch_t* batch, r32* rectangle, r32* uvs, vec4* colors, vec4 color, u32 texture_handle, u32 bytes_per_pixel ){
    
    _assert( rectangle );
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_triangle_strip_textured;
        
        u32 triangle_count = 2;
        u8 vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
        umm array_size = ( 2 + triangle_count ) * vertex_size * sizeof( r32 );
        result->triangle_strip_textured.vertex_array = ( r32* ) memory_push_size( &batch->datas, array_size, 4 );
        _assert( result->triangle_strip_textured.vertex_array );
        result->triangle_strip_textured.triangle_count = triangle_count;
        result->triangle_strip_textured.texture = texture_handle;
        result->triangle_strip_textured.bytes_per_pixel = bytes_per_pixel;
        
        vec4 dummy_colors[ 4 ] = { white, white, white, white };
        r32 dummy_uvs[ 4 ] = { 0, 0, 1, 1 };
        
        r32* va = result->triangle_strip_textured.vertex_array;
        u32* va_u32 = ( u32* ) va;
        
        b32 has_color = gl_vertex_format_has_color( batch->vertex_format );
        b32 has_uv = gl_vertex_format_has_uv( batch->vertex_format );
        
        if ( !colors ) {
            colors = dummy_colors;
        }
        
        if ( !uvs ) {
            uvs = dummy_uvs;
        }
        
        {
            umm index = 0;
            va[ index + 0 ] = rectangle[ 0 ];
            va[ index + 1 ] = rectangle[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 0 ];
            va[ index + 1 ] = rectangle[ 1 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 2 ];
            va[ index + 1 ] = rectangle[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = rectangle[ 2 ];
            va[ index + 1 ] = rectangle[ 1 ];
        }
        
        if ( has_color ) {
            umm index = 2;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 2 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 0 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 3 ], color ) );
            index += vertex_size;
            va_u32[ index ] = v4_to_u32_color( v4_hadamard( colors[ 1 ], color ) );
        }
        
        if ( has_uv ) {
            umm index = 2;
            if ( has_color ) {
                index = 3;
            }
            va[ index + 0 ] = uvs[ 0 ];
            va[ index + 1 ] = uvs[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 0 ];
            va[ index + 1 ] = uvs[ 1 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 2 ];
            va[ index + 1 ] = uvs[ 3 ];
            index += vertex_size;
            va[ index + 0 ] = uvs[ 2 ];
            va[ index + 1 ] = uvs[ 1 ];
        }
    }
    
    return result;
}

stu gl_command_t* gl_batch_triangle_strip( gl_batch_t* batch, r32* vertex_array, u32 triangle_count, vec4 color ) {
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_triangle_strip;
        
        u8 vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
        umm array_size = ( 2 + triangle_count ) * vertex_size * sizeof( r32 );
        result->triangle_strip.vertex_array = ( r32* ) memory_push_copy_p( &batch->datas, vertex_array, array_size, 4 );
        _assert( result->triangle_strip.vertex_array );
        
        result->triangle_strip.triangle_count = triangle_count;
        result->triangle_strip.color = v4_to_u32_color( color );
    }
    
    return result;
}

stu gl_command_t* gl_batch_triangle_strip_textured( gl_batch_t* batch, r32* vertex_array, u32 triangle_count, vec4 color, u32 texture, u32 bytes_per_pixel ) {
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_triangle_strip_textured;
        
        u8 vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
        umm array_size = ( 2 + triangle_count ) * vertex_size * sizeof( r32 );
        result->triangle_strip_textured.vertex_array = memory_push_copy_p( &batch->datas, vertex_array, array_size, 4 );
        _assert( result->triangle_strip_textured.vertex_array );
        
        result->triangle_strip_textured.triangle_count = triangle_count;
        result->triangle_strip_textured.color = v4_to_u32_color( color );
        result->triangle_strip_textured.texture = texture;
        result->triangle_strip_textured.bytes_per_pixel = bytes_per_pixel;
    }
    
    return result;
}

stu gl_command_t* gl_batch_text( gl_batch_t* batch, void* string, umm byte_count, umm codepoint_count, r32 x, r32 baseline_y, b32 do_first_line_character_offset, u32 font_index, vec4 color ) {
    
    /* NOTE simon: when calling this function with the result of the layout functions, do_first_line_character_offset should be set to false as the layout functions already take that into account. */
    
    gl_command_t* result = memory_push_struct( &batch->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        batch->command_count++;
        result->type = gl_command_type_text;
        
        result->text.position = v3( x, baseline_y, 0 );
        result->text.string = memory_get_block( &batch->datas, byte_count, 1 );
        string_push_length( &result->text.string, ( char* ) string, byte_count );
        
        if ( codepoint_count == 0 && byte_count != 0 ) {
            result->text.codepoint_count = ( u32 ) utf8_get_codepoint_count( string, ( u8* ) string + byte_count  );
        } else {
            result->text.codepoint_count = ( u32 ) codepoint_count;
        }
        
        result->text.outputed_codepoint_count = 0;
        result->text.font_index = font_index;
        result->text.color = v4_to_u32_color( color );
        result->text.do_first_line_character_offset = do_first_line_character_offset;
    }
    
    return result;
}

stu void gl_change_rectangle_color( gl_batch_t* batch, gl_command_t* command, vec4* colors, vec4 color ) {
    
    b32 has_color = gl_vertex_format_has_color( batch->vertex_format );
    umm vertex_size = gl_vertex_format_compute_vertex_size( batch->vertex_format );
    umm position_component_count = gl_vertex_format_get_position_component_count( batch->vertex_format );
    _assert( has_color );
    
    u32* va_colors = ( u32* ) command->triangle_strip.vertex_array;
    va_colors += position_component_count;
    
    vec4 dummy_colors[ 4 ] = { white, white, white, white };
    
    if ( !colors ) {
        colors = dummy_colors;
    }
    
    for ( umm i = 0; i < 4; i++ ) {
        va_colors[ i * vertex_size ] = v4_to_u32_color( v4_hadamard( colors[ 0 ], color ) );
    }
}

stu gl_command_t* gl_push_batch( gl_command_buffer_t* commands, gl_batch_t* batch ) {
    
    gl_command_t* result = memory_push_struct( &commands->commands, gl_command_t );
    _assert( result );
    
    if ( result ) {
        commands->command_count++;
        result->type = gl_command_type_batch;
        result->batch.batch = batch;
    }
    
    return result;
}

stu gl_batch_t gl_batch_create_point( data_t* memory, umm point_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 position_component_count = gl_vertex_format_get_position_component_count( vertex_format );
    _assert( !gl_vertex_format_has_uv( vertex_format ) );
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    u8 index_size_byte = gl_vertex_format_get_index_size_byte( vertex_format );
    
    gl_batch_t batch = { 0 };
    
    batch.type = gl_command_type_point;
    batch.vertex_format = vertex_format;
    batch.v_buffer_size = ( u32 ) ( vertex_size_byte * point_count * ( ( position_component_count == 2 ) ? 4 : 8 ) );
    batch.i_buffer_size = ( u32 ) ( index_size_byte * point_count * ( ( position_component_count == 2 ) ? 5 : 15 ) );
    batch.shader_handle = shader_handle;
    batch.matrix = matrix;
    
    umm size = sizeof( gl_command_t );
    
    batch.commands = memory_get_block( memory, size * command_count, 8 );
    batch.datas = memory_get_block( memory, vertex_size_byte * point_count, 4 );
    
    return batch;
}

stu gl_batch_t gl_batch_create_line( data_t* memory, umm point_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    _assert( !gl_vertex_format_has_uv( vertex_format ) );
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    u8 index_size_byte = gl_vertex_format_get_index_size_byte( vertex_format );
    
    gl_batch_t batch = { 0 };
    
    batch.type = gl_command_type_line;
    batch.vertex_format = vertex_format;
    batch.v_buffer_size = ( u32 ) ( vertex_size_byte * point_count );
    batch.i_buffer_size = ( u32 ) ( index_size_byte * ( point_count + command_count ) );
    batch.shader_handle = shader_handle;
    batch.matrix = matrix;
    
    umm size = sizeof( gl_command_t );
    
    batch.commands = memory_get_block( memory, size * command_count, 8 );
    batch.datas = memory_get_block( memory, vertex_size_byte * point_count, 4 );
    
    return batch;
}

stu gl_batch_t gl_batch_create_triangle_strip( data_t* memory, umm triangle_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    u8 index_size_byte = gl_vertex_format_get_index_size_byte( vertex_format );
    
    gl_batch_t batch = { 0 };
    
    batch.type = gl_command_type_triangle_strip;
    batch.vertex_format = vertex_format;
    batch.v_buffer_size = ( u32 ) ( vertex_size_byte * vertex_count );
    batch.i_buffer_size = ( u32 ) ( ( vertex_count + command_count ) * index_size_byte );
    batch.shader_handle = shader_handle;
    batch.matrix = matrix;
    
    umm size = sizeof( gl_command_t );
    
    batch.commands = memory_get_block( memory, size * command_count, 8 );
    batch.datas = memory_get_block( memory, vertex_size_byte * vertex_count, 4 );
    
    return batch;
}

stu gl_batch_t gl_batch_create_triangle_strip_textured( data_t* memory, umm triangle_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    u8 index_size_byte = gl_vertex_format_get_index_size_byte( vertex_format );
    
    gl_batch_t batch = { 0 };
    
    batch.type = gl_command_type_triangle_strip_textured;
    batch.vertex_format = vertex_format;
    batch.v_buffer_size = ( u32 ) ( vertex_size_byte * vertex_count );
    batch.i_buffer_size = ( u32 ) ( ( vertex_count + command_count ) * index_size_byte );
    batch.shader_handle = shader_handle;
    batch.matrix = matrix;
    
    umm size = sizeof( gl_command_t );
    
    batch.commands = memory_get_block( memory, size * command_count, 8 );
    batch.datas = memory_get_block( memory, vertex_size_byte * vertex_count, 4 );
    
    return batch;
}

stu gl_batch_t gl_batch_create_text( data_t* memory, umm one_byte_character_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix, font_collection_t* collection ) {
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    u8 index_size_byte = gl_vertex_format_get_index_size_byte( vertex_format );
    
    gl_batch_t batch = { 0 };
    
    batch.type = gl_command_type_text;
    batch.vertex_format = vertex_format;
    batch.v_buffer_size = ( u32 ) ( vertex_size_byte * 4 * one_byte_character_count );
    batch.i_buffer_size = ( u32 ) ( one_byte_character_count * 5 * index_size_byte );
    batch.shader_handle = shader_handle;
    batch.matrix = matrix;
    
    umm size = sizeof( gl_command_t );
    
    batch.commands = memory_get_block( memory, size * command_count, 8 );
    batch.datas = memory_get_block( memory, one_byte_character_count, 1 );
    batch.collection = collection;
    
    return batch;
}

stu void gl_batch_frame_init( gl_batch_t* batch ) {
    
    batch->command_count = 0;
    batch->commands.used = 0;
    batch->datas.used = 0;
    batch->vertex_used = 0;
    batch->vertex_index = 0;
    batch->index_used = 0;
}

stu font_t* gl_batch_get_font( gl_batch_t* batch, u32 font_id ) {
    _assert( batch->type == gl_command_type_text );
    _assert( batch->collection );
    _assert( batch->collection->fonts );
    _assert( batch->collection->font_count > font_id );
    font_t* result = batch->collection->fonts + font_id;
    return result;
}

// MARK render

stu void internal_gl_push_index( void* i_array, umm* i_array_index, umm value, umm index_size ) {
    
    if ( index_size == 1 ) {
        ( ( u8* ) i_array )[ *i_array_index ] = ( u8 ) value;
    } else if ( index_size == 2 ) {
        ( ( u16* ) i_array )[ *i_array_index ] = ( u16 ) value;
    } else {
        ( ( u32* ) i_array )[ *i_array_index ] = ( u32 ) value;
    }
    
    ( *i_array_index )++;
}

stu void internal_gl_push_primitive_restart( void* i_array, umm* i_array_index, umm index_size ) {
    
    if ( index_size == 1 ) {
        ( ( u8* ) i_array )[ *i_array_index ] = gl_primitive_restart_index_u8;
    } else if ( index_size == 2 ) {
        ( ( u16* ) i_array )[ *i_array_index ] = gl_primitive_restart_index_u16;
    } else {
        ( ( u32* ) i_array )[ *i_array_index ] = gl_primitive_restart_index_u32;
    }
    
    ( *i_array_index )++;
}

stu u32 gl_text_to_vertex_array( void* text, umm byte_count, r32 x, r32 y, b32 do_first_line_character_offset, font_collection_t* collection, u32 font_index, r32* v_array, u32* i_array, umm index, u32 color, u8 vertex_format ) {
    
    u8 index_size = gl_vertex_format_get_index_size_byte( vertex_format );
    
    umm vertex_size = gl_vertex_format_compute_vertex_size( vertex_format );
    r32 sheet_width = ( r32 ) collection->atlas_width;
    r32 sheet_height = ( r32 ) collection->atlas_height;
    
    font_t* font = collection->fonts + font_index;
    r32 scale = font->scale;
    r32 line_start_x = x;
    
    b32 line_first_character = true;
    umm index_offset = 0;
    u32 outputed_codepoint_count = 0;
    u8* current = ( u8* ) text;
    u8* end = current + byte_count;
    
    while ( current < end ) {
        
        u32 codepoint = *current;
        u8 codeunit_count = 0;
        
        if ( codepoint == '\r' ) {
            
            codeunit_count = 1;
            
            if ( current + 1 >= end || *( current + 1 ) != '\n' ) {
                x = line_start_x;
                y += ( font->descent - font->line_gap - font->ascent ) * scale;
                line_first_character = true;
            }
            
        } else if ( codepoint == '\n' ) {
            
            codeunit_count = 1;
            x = line_start_x;
            y += ( font->descent - font->line_gap - font->ascent ) * scale;
            line_first_character = true;
            
        } else {
            
            u32 error = 0;
            utf8_decode( ( u8* ) current, end, &codepoint, &codeunit_count, &error );
            
            if ( is_error( &error ) ) {
                /* NOTE simon: if the user passes an invalid utf8 string we stop. */
                break;
            }
            
            outputed_codepoint_count++;
            
            u32 metrics_index = font_get_codepoint_metrics_index( font, codepoint );
            codepoint_metrics_t* cm = font->cp_metrics + metrics_index;
            
            r32 x_offset = 0;
            r32 advance = cm->advance;
            
            if ( line_first_character && do_first_line_character_offset ) {
                r32 offset = font_first_character_offset( font, cm );
                x_offset = -offset;
                advance -= offset;
                line_first_character = false;
            }
            
            if ( current + codeunit_count < end ) {
                
                u32 next_codepoint = 0;
                u8 next_codeunit_count = 0;
                utf8_decode( ( u8* ) current + codeunit_count, end, &next_codepoint, &next_codeunit_count, &error );
                
                if ( is_error( &error ) ) {
                    next_codepoint = ucs_replacement_codepoint;
                    next_codeunit_count = 1;
                }
                r32 kerning = font_get_kerning( font, cm, next_codepoint ) * scale;
                advance += kerning;
            }
            
            umm v_array_index = 0;
            u32* c_array = ( u32* ) v_array;
            
            v_array[ v_array_index++ ] = x + cm->x0 + x_offset;
            v_array[ v_array_index++ ] = y + cm->y0;
            c_array[ v_array_index++ ] = color;
            v_array[ v_array_index++ ] = ( r32 ) cm->u0 / sheet_width;
            v_array[ v_array_index++ ] = ( r32 ) cm->v0 / sheet_height;
            
            v_array[ v_array_index++ ] = x + cm->x1 + x_offset;
            v_array[ v_array_index++ ] = y + cm->y0;
            c_array[ v_array_index++ ] = color;
            v_array[ v_array_index++ ] = ( r32 ) cm->u1 / sheet_width;
            v_array[ v_array_index++ ] = ( r32 ) cm->v0 / sheet_height;
            
            v_array[ v_array_index++ ] = x + cm->x0 + x_offset;
            v_array[ v_array_index++ ] = y + cm->y1;
            c_array[ v_array_index++ ] = color;
            v_array[ v_array_index++ ] = ( r32 ) cm->u0 / sheet_width;
            v_array[ v_array_index++ ] = ( r32 ) cm->v1 / sheet_height;
            
            v_array[ v_array_index++ ] = x + cm->x1 + x_offset;
            v_array[ v_array_index++ ] = y + cm->y1;
            c_array[ v_array_index++ ] = color;
            v_array[ v_array_index++ ] = ( r32 ) cm->u1 / sheet_width;
            v_array[ v_array_index++ ] = ( r32 ) cm->v1 / sheet_height;
            
            internal_gl_push_index( i_array, &index_offset, index + 0, index_size );
            internal_gl_push_index( i_array, &index_offset, index + 1, index_size );
            internal_gl_push_index( i_array, &index_offset, index + 2, index_size );
            internal_gl_push_index( i_array, &index_offset, index + 3, index_size );
            internal_gl_push_primitive_restart( i_array, &index_offset, index_size );
            
            index += 4;
            v_array += ( vertex_size * 4 );
            
            x += advance;
        }
        
        current += codeunit_count;
    }
    
    return outputed_codepoint_count;
}

#define GL_H

#endif

