
#ifndef IMAGE_H

typedef enum image_resize_error_t {
    image_resize_error_none,
    image_resize_error_width_and_height_are_zero,
    image_resize_error_component_count_is_invalid,
    image_resize_error_count
} image_resize_error_t;

void image_resize( void* src_arg, u32 src_width, u32 src_height, void* dst_arg, u32* dst_width, u32* dst_height, umm component_count, b32 flip_r_and_b, r32 gamma, u32* error ) {
    
    profiler_event_start( image_resize );
    
    if ( no_error( error ) ) {
        
        u8* src = ( u8* ) src_arg;
        u8* dst = ( u8* ) dst_arg;
        
        r32 src_u_increment = 1.0f / ( r32 ) src_width;
        r32 src_v_increment = 1.0f / ( r32 ) src_height;
        r32 dst_u_increment = 1.0f / ( r32 ) ( *dst_width );
        r32 dst_v_increment = 1.0f / ( r32 ) ( *dst_height );
        
        if ( ( *dst_width ) ) {
            r32 height = ( r32 ) src_height * ( ( r32 ) ( *dst_width ) / ( r32 ) src_width );
            *dst_height = ( u32 ) height;
            r32 src_v_max = ( r32 ) ( *dst_height ) / height; /* NOTE simon: since the we cut a pixel at the top of the image, we remove that v part from the source. */
            dst_v_increment = src_v_max / ( r32 ) ( *dst_height );
        } else if ( *dst_height ) {
            r32 width = ( r32 ) src_width * ( ( r32 ) ( *dst_height ) / ( r32 ) src_height );
            *dst_width = ( u32 ) width;
            r32 src_u_max = ( r32 ) ( *dst_width ) / width;
            dst_u_increment = src_u_max / ( r32 ) ( *dst_width );
        } else {
            set_error( error, image_resize_error_width_and_height_are_zero );
        }
        
        if ( no_error( error ) && src && dst && ( component_count == 0 || component_count > 4 ) ) {
            set_error( error, image_resize_error_component_count_is_invalid );
        }
        
        if ( no_error( error ) && src && dst ) {
            
            r32 dst_v_min = 0;
            umm dst_base = 0;
            
            for ( umm ty = 0; ty < ( *dst_height ); ty++ ) {
                
                r32 dst_v_max = ( ty + 1 ) * dst_v_increment;
                r32 dst_u_min = 0;
                
                for ( umm tx = 0; tx < ( *dst_width ); tx++ ) {
                    
                    r32 dst_u_max = ( tx + 1 ) * dst_u_increment;
                    
                    r32 dst_r32[ 4 ] = { 0 };
                    umm y = ( umm ) ( src_height * dst_v_min );
                    r32 src_v_min = y * src_v_increment;
                    r32 src_v_max = ( y + 1 ) * src_v_increment;
                    
                    while ( src_v_min < dst_v_max && y < src_height ) {
                        
                        umm x = ( umm ) ( src_width * dst_u_min );
                        r32 src_u_min = x * src_u_increment;
                        r32 src_u_max = ( x + 1 ) * src_u_increment;
                        
                        umm base = ( y * src_width + x ) * component_count;
                        
                        r32 min_v = math_max( src_v_min, dst_v_min );
                        r32 max_v = math_min( src_v_max, dst_v_max );
                        r32 v_factor = ( max_v - min_v ) / dst_v_increment;
                        
                        while ( src_u_min < dst_u_max && x < src_width ) {
                            
                            r32 min_u = math_max( src_u_min, dst_u_min );
                            r32 max_u = math_min( src_u_max, dst_u_max );
                            r32 u_factor = ( max_u - min_u ) / dst_u_increment;
                            
                            r32 factor = v_factor * u_factor;
                            
                            for ( umm i = 0; i < component_count; i++ ) {
                                dst_r32[ i ] += src[ base + i ] * factor;
                            }
                            
                            base += component_count;
                            x++;
                            src_u_min = src_u_max;
                            src_u_max = ( x + 1 ) * src_u_increment;
                        }
                        
                        y++;
                        src_v_min = src_v_max;
                        src_v_max = ( y + 1 ) * src_v_increment;
                    }
                    
                    if ( gamma != 1.0f ) {
                        for ( umm i = 0; i < component_count; i++ ) {
                            dst_r32[ i ] = expf( gamma * logf( dst_r32[ i ] / 255.0f ) );
                            dst[ dst_base + i ] = ( u8 ) roundf( dst_r32[ i ] * 255.0f );
                        }
                    } else {
                        for ( umm i = 0; i < component_count; i++ ) {
                            dst[ dst_base + i ] = ( u8 ) roundf( dst_r32[ i ] );
                        }
                    }
                    
                    if ( component_count > 2 && flip_r_and_b ) {
                        u8 temp = dst[ dst_base + 0 ];
                        dst[ dst_base + 0 ] = dst[ dst_base + 2 ];
                        dst[ dst_base + 2 ] = temp;
                    }
                    
                    dst_base += component_count;
                    dst_u_min = dst_u_max;
                }
                
                dst_v_min = dst_v_max;
            }
        }
    }
    
    profiler_event_end( image_resize );
}

stu void image_copy_sub_data( u8* source, u8* destination, s32 width, s32 height, u8 bytes_per_pixel, s32 destination_width ) {
    
    s32 source_pitch = width * bytes_per_pixel;
    s32 destination_pitch = destination_width * bytes_per_pixel;
    
    for ( s32 y = 0; y < height; y++ ) {
        
        u8* d = destination;
        u8* s = source;
        
        for ( s32 x = 0; x < width; x++ ) {
            for ( s32 i = 0; i < bytes_per_pixel; i++ ) {
                *d++ = *s++;
            }
        }
        
        destination += destination_pitch;
        source += source_pitch;
    }
}

stu void image_rgba_to_grayscale( u8* source, u8* destination, s32 width, s32 height, s32 pitch ) {
    
    if ( pitch == 0 ) {
        pitch = width;
    }
    
    for ( s32 y = 0; y < height; y++ ) {
        
        u8* s = source;
        u8* d = destination;
        
        for ( s32 x = 0; x < width; x++ ) {
            
            u8 value = ( ( u32 ) s[ 0 ] + ( u32 ) s[ 1 ] + ( u32 ) s[ 2 ] ) / 3;
            d[ 0 ] = value;
            d++;
            s += 4;
        }
        
        source += pitch * 4;
        destination += pitch;
    }
}

typedef struct image_to_ascii_t {
    u8* atlas;
    u16 width;
    u16 height;
    u16 pitch;
    u8 first_element;
    u8 last_element;
    u8 x_element_count;
    u8 y_element_count;
    u8 element_width;
    u8 element_height;
    s32 intensity[ 256 ];
    s32 max_intensity;
} image_to_ascii_t;

stu void image_to_ascii_initialize( image_to_ascii_t* ascii ) {
    
    ascii->max_intensity = 0;
    
    for ( umm i = ascii->first_element; i <= ascii->last_element; i++ ) {
        
        ascii->intensity[ i ] = 0;
        
        umm cx = i % ascii->x_element_count;
        umm cy = i / ascii->x_element_count;
        
        u8* pixels = ascii->atlas
            + cy * ascii->pitch * ascii->element_height
            + cx * ascii->element_width;
        
        for ( umm y = 0; y < ascii->element_height; y++ ) {
            
            for ( umm x = 0; x < ascii->element_width; x++ ) {
                ascii->intensity[ i ] += pixels[ x ];
            }
            
            pixels += ascii->pitch;
        }
        
        if ( ascii->intensity[ i ] > ascii->max_intensity ) {
            ascii->max_intensity = ascii->intensity[ i ];
        }
    }
    
    _assert( ascii->max_intensity );
    
    s32 absolute_max_intensity = ascii->element_width * ascii->element_height * 255;
    
    for ( umm i = ascii->first_element; i <= ascii->last_element; i++ ) {
        ascii->intensity[ i ] = ( s32 ) ( ( ( r32 ) ascii->intensity[ i ] / ( r32 ) ascii->max_intensity ) * absolute_max_intensity );
    }
}

b32 g_extra_step = false;

stu void image_to_ascii_render( u8* pixels, umm width, umm height, umm pitch, image_to_ascii_t* ascii, u8* ascii_out, umm ascii_count, umm* out_ascii_x_count, umm* out_ascii_y_count ) {
    
    profiler_event_start( image_to_ascii_render );
    
    umm x_element_count = width / ascii->element_width;
    umm y_element_count = height / ascii->element_height;
    
    *out_ascii_x_count = x_element_count;
    *out_ascii_y_count = y_element_count;
    
    if ( x_element_count * y_element_count <= ascii_count ) {
        
        for ( umm y = 0; y < y_element_count; y++ ) {
            
            for ( umm x = 0; x < x_element_count; x++ ) {
                
                s32 intensity = 0;
                u8* line = pixels + y * pitch * ascii->element_height + x * ascii->element_width;
                
                for ( umm cy = 0; cy < ascii->element_height; cy++ ) {
                    
                    for ( umm cx = 0; cx < ascii->element_width; cx++ ) {
                        intensity += line[ cx ];
                    }
                    
                    line += pitch;
                }
                
                s32 best_difference = s32_max;
                umm best = 0;
                
                for ( umm i = ascii->first_element; i <= ascii->last_element; i++ ) {
                    
                    s32 intensity_difference = intensity - ascii->intensity[ i ];
                    
                    if ( intensity_difference < 0 ) {
                        intensity_difference *= -1;
                    }
                    
                    if ( intensity_difference < best_difference ) {
                        best_difference = intensity_difference;
                        best = i;
                    } else if ( intensity_difference == best_difference && g_extra_step ) {
                        
                        profiler_event_start( extra_step );
                        
                        umm cx_new = i % ascii->x_element_count;
                        umm cy_new = i / ascii->x_element_count;
                        umm cx_best = best % ascii->x_element_count;
                        umm cy_best = best / ascii->x_element_count;
                        
                        u8* c_new = ascii->atlas + cy_new * ascii->pitch * ascii->element_height + cx_new * ascii->element_width;
                        u8* c_best = ascii->atlas + cy_best * ascii->pitch * ascii->element_height + cx_best * ascii->element_width;
                        u8* ref = pixels + y * pitch * ascii->element_height + x * ascii->element_width;
                        u8 score_new = 0;
                        u8 score_best = 0;
                        
                        for ( umm cy = 0; cy < ascii->element_height; cy++ ) {
                            for ( umm cx = 0; cx < ascii->element_width; cx++ ) {
                                u8 v_new = c_new[ cx ] == 255;
                                u8 v_best = c_best[ cx ] == 255;
                                u8 v_ref = ref[ cx ] > 127;
                                
                                if ( v_new == v_ref ) {
                                    score_new++;
                                }
                                
                                if ( v_best == v_ref ) {
                                    score_best++;
                                }
                            }
                            
                            c_new += ascii->pitch;
                            c_best += ascii->pitch;
                            ref += pitch;
                        }
                        
                        if ( score_new > score_best ) {
                            best = i;
                        }
                        
                        profiler_event_end( extra_step );
                    }
                }
                
                ascii_out[ y * x_element_count + x ] = ( u8 ) best;
            }
        }
    }
    
    profiler_event_end( image_to_ascii_render );
}

#define IMAGE_H 1

#endif
