
#ifndef SOUND_H

enum sound_channel_t {
    
    sound_channel_front_left = 0x1 << 0,
    sound_channel_front_right = 0x1 << 1,
    sound_channel_front_center = 0x1 << 2,
    sound_channel_low_frequency = 0x1 << 3,
    sound_channel_back_left = 0x1 << 4,
    sound_channel_back_right = 0x1 << 5,
    sound_channel_front_left_of_center = 0x1 << 6,
    sound_channel_front_right_of_center = 0x1 << 7,
    sound_channel_back_center = 0x1 << 8,
    sound_channel_side_left = 0x1 << 9,
    sound_channel_side_right = 0x1 << 10,
    sound_channel_top_center = 0x1 << 11,
    sound_channel_top_front_left = 0x1 << 12,
    sound_channel_top_front_center = 0x1 << 13,
    sound_channel_top_front_right = 0x1 << 14,
    sound_channel_top_back_left = 0x1 << 15,
    sound_channel_top_back_center = 0x1 << 16,
    sound_channel_top_back_right = 0x1 << 17,
};

typedef struct sound_format_t {
    
    u32 samples_per_second;
    u16 bits_per_sample;
    u16 channel_count;
    u32 channel_mask;
    u16 container_size;
} sound_format_t;

stu void sound_format_initialize( sound_format_t* format, u32 samples_per_seconds, u16 bits_per_sample, u16 channel_count, u32 channel_mask ) {
    
    format->samples_per_second = 48000;
    format->bits_per_sample = 24;
    format->channel_count = 2;
    format->channel_mask = sound_channel_front_left | sound_channel_front_right;
    b32 bit_remaining = ( format->bits_per_sample % 8 ) != 0;
    u16 container_size = ( format->bits_per_sample / 8 ) + ( ( bit_remaining ) ? 1 : 0 );
    format->container_size = container_size;
}

stu u32 sound_compute_buffer_size( sound_format_t* format, r32 length ) {
    
    u32 length_in_samples = ( u32 ) ceil( format->samples_per_second * length );
    u32 data_size = format->container_size * format->channel_count * length_in_samples;
    return data_size;
}

typedef struct sound_sound_t {
    
    u32 current_sample;
    u32 sample_count;
    data_t* memory;
    
} sound_sound_t;

#define SOUND_COUNT 128

typedef struct sound_mixer_t {
    
    sound_format_t format;
    sound_sound_t sounds[ SOUND_COUNT ];
    u32 sound_free_list[ SOUND_COUNT ];
    u32 sound_free_list_count;
    
    data_t buffer;
} sound_mixer_t;

#undef SOUND_COUNT

stu void sound_mixer_initialize( data_t* memory, sound_mixer_t* mixer, sound_format_t* format, u32 buffer_size_in_samples ) {
    
    umm buffer_size_in_bytes = buffer_size_in_samples * format->channel_count * format->container_size;
    mixer->format = *format;
    
    for ( umm sound_index = 0; sound_index < array_count( mixer->sounds ); sound_index++ ) {
        
        mixer->sounds[ sound_index ].current_sample = 0;
        mixer->sounds[ sound_index ].sample_count = 0;
        mixer->sounds[ sound_index ].memory = 0;
        mixer->sound_free_list[ sound_index ] = ( u32 ) sound_index;
    }
    
    mixer->sound_free_list_count = array_count( mixer->sounds );
    mixer->buffer = memory_get_block( memory, buffer_size_in_bytes, 0 );
}

stu void sound_play( sound_mixer_t* mixer, data_t* sound, u32 sample_count ) {
    
    if ( mixer->sound_free_list_count > 0 ) {
        mixer->sound_free_list_count--;
        sound_sound_t* s = mixer->sounds + mixer->sound_free_list[ mixer->sound_free_list_count ];
        s->memory = sound;
        s->current_sample = 0;
        s->sample_count = sample_count;
    }
}

stu void sound_mixer_render( sound_mixer_t* mixer, u32 sample_count ) {
    
    u32 container_size = mixer->format.container_size;
    u32 channel_count = mixer->format.channel_count;
    u32 bytes_per_sample = channel_count * container_size;
    memory_zero( mixer->buffer.bytes, sample_count * bytes_per_sample  );
    
    for ( umm sound_index = 0; sound_index < array_count( mixer->sounds ); sound_index++ ) {
        
        sound_sound_t* sound = mixer->sounds + sound_index;
        
        if ( sound->sample_count > 0 ) {
            
            u32 left_to_write = sound->sample_count - sound->current_sample;
            u32 samples_to_write = ( left_to_write > sample_count ) ? sample_count : left_to_write;
            
            u8* destination = mixer->buffer.bytes;
            u8* source = sound->memory->bytes + ( sound->current_sample * bytes_per_sample );
            
            for ( umm sample_index = 0; sample_index < samples_to_write; sample_index++ )  {
                
                for ( umm channel_index = 0; channel_index < channel_count; channel_index++ ) {
                    
                    if ( container_size == 1 ) {
                        
                        *destination += *source;
                        destination++;
                        source++;
                        
                    } else if ( container_size == 2 )  {
                        
                        *( ( u16* ) destination ) += *( ( s16* ) source );
                        destination += 2;
                        source += 2;
                        
                    } else if ( container_size == 3 ) {
                        
                        s32 s = math_s24_to_s32( source );
                        s32 d = math_s24_to_s32( destination );
                        s32 output = s + d;
                        
                        s32 o = math_s32_to_s24( output );
                        u8* o_bytes = ( u8* ) &o;
                        
                        destination[ 0 ] = o_bytes[ 0 ];
                        destination[ 1 ] = o_bytes[ 1 ];
                        destination[ 2 ] = o_bytes[ 2 ];
                        
                        source += 3;
                        destination += 3;
                        
                    } else {
                        
                        _assert( container_size == 4 );
                        
                        *( ( s32* ) destination ) += *( ( s32* ) source );
                        destination += 4;
                        source += 4;
                    }
                }
            }
            
            sound->current_sample += samples_to_write;
            
            if ( sound->current_sample >= sound->sample_count ) {
                
                sound->sample_count = 0;
                mixer->sound_free_list[ mixer->sound_free_list_count ] = ( u32 ) sound_index;
                mixer->sound_free_list_count++;
            }
        }
    }
}

#define SOUND_H

#endif
